module reg_file(a1, a2, a3, R1, R2, R3, clk, reg_write);
   parameter width = 32, prop = 0;  
   input [4:0] a1;               // A register address to read from.
   input [4:0] a2;               // A register address to read from.
   input [4:0] a3;               // The register address to write to.
   input [width-1:0] R3;         // The value to write to a3.
   input 	     clk;        // The clock.
   input 	     reg_write;  // If reg_write == 1'b1 then write to a3.
   output [width-1:0] R1;        // The contents of register address a1.
   output [width-1:0] R2;        // The contents of register address a2.
   
   reg [width-1:0]    R1, R2;
   reg [width-1:0]    register_file [0:31];
   reg [width-1:0]    tmp1;
   reg [width-1:0]    tmp2;
   reg [width-1:0]    tmp3,tmp4;
   
`define	TRACE_REGS	1	// Trace the contents of all registers.
      
   always
     begin 
	#(prop)R1 = register_file[a1];
	R2 = register_file[a2];      	
     end
   
   always @(negedge clk)             
     begin
	if(reg_write == 1'b1)
	  begin	
             #(prop) register_file[a3] = R3; 		
	  end
	print_trace;	
     end
   
   task print_trace;
      integer i;
      integer j;
      integer k;
      
      begin
`ifdef TRACE_REGS
	 begin
            k = 0;
	    $display("Simulation time is %d",$time);
            for (i = 0; i < 32; i = i + 4) 
              begin
		 $write("R[%d]: ",k);
		 for (j = 0; j <= 3; j = j + 1) 
		   begin
		      $write("  %h",register_file[k]);
		      k = k + 1;
		   end
		 $write("\n");
              end
	    $write("\n");
	 end
`endif	// TRACE_REGS
      end
   endtask // print_trace        
endmodule // reg_file









