/*********************************************************************

        This module is for stage1 of the processor.
        The following are the modules which are instantiated.
        
        1. selIR2mux --> To select IR2 input
        2. pcmux     --> To select Program counter input
        3. pc2mux    --> To select PC2 input
        4. inc4      --> Module to increment PC by 4
        5. d_reg     --> Registers to latch values for stage 2 at clock.

**********************************************************************/
module stage1(ir2, pc2, pcout, ir1, z4, mp1, clk, clr, pause1, pause2, nop2);   
   input [31:0]  ir1;    // The instruction read from memory. 
   input [31:0]  z4;     // The address calculated in stage 3. 
   input 	 mp1;    // If mp == 1 then there is nither a jump nor a taken conditional branch in pipeline stage 4.
   input 	 clk;    // The clock. 
   input 	 clr;    // The clear signal used to reset all D-type flip-flops. 
   input 	 pause1; // This is identical to pause2. Why?
   input 	 pause2; // Pause the pipeline processing for this stage.
   input 	 nop2;   // Perform a NOP.
   output [31:0] ir2;    // The Instruction Register we will pass on to pipeline stage 2.
   output [31:0] pc2;    // The Program Counter we will pass on to pipeline stage 2.
   output [31:0] pcout;  // The Program Counter sent to memory which is used to read ir1.
   
   // defining internal wires
   wire [31:0] 	 pcin;
   wire [31:0] 	 pc2_in;
   wire [31:0] 	 inc4out;
   wire [31:0] 	 ir1_mux;   
   wire [31:0] 	 nop2_out;           
   wire 	 or_nop2_pause1;
   wire 	 sel1;
      
   or #3 (or_nop2_pause1, pause1, nop2);
   and #3 (sel1, mp1, or_nop2_pause1);
         
   // paras are sel0, sel1, out, in0-3
   mux4 #(32,6) selIR2mux(nop2, pause2, ir1_mux, ir1, nop2_out, ir2, ir2);
   mux4 #(32,6) pcmux(mp1, sel1, pcin, z4, inc4out, pcout, pcout);
   mux2 #(32,6) pc2mux(pause2, pc2_in, inc4out, pc2);	
   inc4 incr(inc4out, pcout);                      // This is a 32-bit machine with one instruction per machine word.
   d_reg #(32,8) pc(clk, pcout, pcin, clr);
   d_reg #(32,8) pc2reg(clk, pc2, pc2_in, clr);
   d_reg #(32,8) ir2reg(clk, ir2, ir1_mux, clr);
   d_reg #(32,8) nop2reg(clk, nop2_out, 0, clr);   // A NOP instruction.
endmodule // stage1
